#![allow(clippy::let_and_return)]
#![forbid(elided_lifetimes_in_paths, unsafe_code)]

use aoc23::read;
use chumsky::{prelude::*, text::int};
use itertools::Itertools as _;
use std::{
	fmt::{self, Debug, Display},
	ops::{Add, Deref, DerefMut}
};

trait Newtype:
	Copy
	+ Debug
	+ Ord
	+ From<usize>
	+ Into<usize>
	+ Deref<Target = usize>
	+ DerefMut
	+ Display
	+ Add<usize, Output = Self>
{
}

macro_rules! newtype {
	($( struct $newtype:ident($inner:ident);)*) => {
		$(
			#[derive(Clone, Copy, Debug, Eq, PartialEq, PartialOrd, Ord)]
			struct $newtype($inner);

			impl From<$inner> for $newtype {
				fn from(inner: $inner) -> Self {
					Self(inner)
				}
			}

			impl From<$newtype> for $inner {
				fn from(newtype: $newtype) -> Self {
					newtype.0
				}
			}

			impl Deref for $newtype {
				type Target = $inner;

				fn deref(&self) -> &$inner {
					&self.0
				}
			}

			impl DerefMut for $newtype {
				fn deref_mut(&mut self) -> &mut $inner {
					&mut self.0
				}
			}

			impl Display for $newtype {
				fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
					write!(f, "{}", self.0)
				}
			}

			impl Add<$inner> for $newtype {
				type Output = $newtype;

				fn add(self, rhs: $inner) -> Self {
					Self(self.0 + rhs)
				}
			}

			impl Newtype for $newtype {}
		)*
	};
}

newtype! {
	struct Seed(usize);
	struct Soil(usize);
	struct Fertilizer(usize);
	struct Water(usize);
	struct Light(usize);
	struct Temperature(usize);
	struct Humidity(usize);
	struct Location(usize);
}

#[derive(Debug)]
struct Range<S, T> {
	source_start: S,
	target_start: T,
	len: usize
}

struct Map<S, T> {
	ranges: Vec<Range<S, T>>
}

impl<S, T> Map<S, T>
where
	S: Newtype,
	T: Newtype
{
	fn get(&self, source: S) -> T {
		for r in &self.ranges {
			let Some(offset) = source.checked_sub(r.source_start.into()) else {
				continue;
			};
			if offset < r.len {
				return (r.target_start.into() + offset).into();
			}
		}

		source.into().into()
	}

	fn get_range_impl(&self, start: S, len: usize) -> Vec<(T, usize)> {
		let mut result = Vec::new();
		for r in &self.ranges {
			let possible_start = start.max(r.source_start);
			let possible_end = (start + len).min(r.source_start + r.len);
			if possible_start < possible_end {
				let range = (
					r.target_start + (possible_start.into() - r.source_start.into()),
					possible_end.into() - possible_start.into()
				);
				// eprintln!("({start:?}, {len}) => {range:?} ({r:?}");
				result.push(range);

				if start < possible_start {
					result.extend(
						self.get_range_impl(start, possible_start.into() - start.into())
					);
				}
				if (start + len) > possible_end {
					result.extend(self.get_range_impl(
						possible_end,
						start.into() + len - possible_end.into()
					));
				}
			}
		}

		if result.is_empty() {
			result.push((start.into().into(), len));
		}

		result
	}

	fn get_range(&self, (start, len): (S, usize)) -> Vec<(T, usize)> {
		let result = self.get_range_impl(start, len);

		// for range in &result {
		// 	eprintln!("({start:?}, {len}) => {range:?}");
		// }

		result
	}

	fn parser() -> impl Parser<char, Self, Error = Simple<char>> {
		int(10)
			.separated_by(just(" "))
			.exactly(3)
			.then_ignore(just("\n"))
			.repeated()
			.map(|ranges| Self {
				ranges: ranges
					.into_iter()
					.map(|ints: Vec<String>| Range {
						source_start: ints[1].parse::<usize>().unwrap().into(),
						target_start: ints[0].parse::<usize>().unwrap().into(),
						len: ints[2].parse().unwrap()
					})
					.collect()
			})
	}
}

struct Almanac {
	seeds: Vec<Seed>,
	seed_to_soil: Map<Seed, Soil>,
	soil_to_fertilizer: Map<Soil, Fertilizer>,
	fertilizer_to_water: Map<Fertilizer, Water>,
	water_to_light: Map<Water, Light>,
	light_to_temperature: Map<Light, Temperature>,
	temperature_to_humidity: Map<Temperature, Humidity>,
	humidity_to_location: Map<Humidity, Location>
}

impl Almanac {
	fn parser() -> impl Parser<char, Self, Error = Simple<char>> {
		fn parse_another_map<'a, P, R, S, T>(
			parser: P,
			name: &'a str
		) -> impl Parser<char, (R, Map<S, T>), Error = Simple<char>> + 'a
		where
			P: Parser<char, R, Error = Simple<char>> + 'a,
			R: 'a,
			S: Newtype + 'a,
			T: Newtype + 'a
		{
			parser
				.then_ignore(just(name))
				.then_ignore(just(" map:\n"))
				.then(Map::parser())
		}

		let parser = just("seeds: ")
			.ignore_then(int(10).separated_by(just(" ")))
			.map(|ints: Vec<String>| {
				ints.into_iter()
					.map(|int| int.parse::<usize>().unwrap().into())
					.collect::<Vec<Seed>>()
			})
			.then_ignore(just("\n\n"));
		let parser = parse_another_map(parser, "seed-to-soil").then_ignore(just("\n"));
		let parser =
			parse_another_map(parser, "soil-to-fertilizer").then_ignore(just("\n"));
		let parser =
			parse_another_map(parser, "fertilizer-to-water").then_ignore(just("\n"));
		let parser = parse_another_map(parser, "water-to-light").then_ignore(just("\n"));
		let parser =
			parse_another_map(parser, "light-to-temperature").then_ignore(just("\n"));
		let parser =
			parse_another_map(parser, "temperature-to-humidity").then_ignore(just("\n"));
		let parser = parse_another_map(parser, "humidity-to-location");

		parser.map(
			|(
				(
					(
						(
							(
								((seeds, seed_to_soil), soil_to_fertilizer),
								fertilizer_to_water
							),
							water_to_light
						),
						light_to_temperature
					),
					temperature_to_humidity
				),
				humidity_to_location
			)| {
				Almanac {
					seeds,
					seed_to_soil,
					soil_to_fertilizer,
					fertilizer_to_water,
					water_to_light,
					light_to_temperature,
					temperature_to_humidity,
					humidity_to_location
				}
			}
		)
	}
}

fn parser() -> impl Parser<char, Almanac, Error = Simple<char>> {
	Almanac::parser().then_ignore(end())
}

fn main() -> anyhow::Result<()> {
	let almanac = read("inputs/day5.txt", parser())?;

	let min_loc = almanac
		.seeds
		.iter()
		.copied()
		.map(|seed| {
			let soil = almanac.seed_to_soil.get(seed);
			let fertilizer = almanac.soil_to_fertilizer.get(soil);
			let water = almanac.fertilizer_to_water.get(fertilizer);
			let light = almanac.water_to_light.get(water);
			let temperature = almanac.light_to_temperature.get(light);
			let humidity = almanac.temperature_to_humidity.get(temperature);
			let location = almanac.humidity_to_location.get(humidity);
			// eprintln!("{seed}, {soil}, {fertilizer}, {water}, {light}, {temperature}, {humidity}, {location}");
			location
		})
		.min()
		.unwrap();
	println!("{min_loc}");

	let min_loc = almanac
		.seeds
		.iter()
		.copied()
		.tuples()
		.map(|(start, len)| (start, len.into()))
		.flat_map(|range| almanac.seed_to_soil.get_range(range))
		.flat_map(|range| almanac.soil_to_fertilizer.get_range(range))
		.flat_map(|range| almanac.fertilizer_to_water.get_range(range))
		.flat_map(|range| almanac.water_to_light.get_range(range))
		.flat_map(|range| almanac.light_to_temperature.get_range(range))
		.flat_map(|range| almanac.temperature_to_humidity.get_range(range))
		.flat_map(|range| almanac.humidity_to_location.get_range(range))
		.map(|(start, _)| start)
		.min()
		.unwrap();
	println!("{min_loc}");

	Ok(())
}
